# -*- coding: utf-8 -*-
"""
Created on Fri Jan 18 14:19:03 2019

@author: Guest Group
"""

__all__ = ('FreeProperty',)

# %% Metaclasses

# Do not include: '__new__', '__init__', '__del__', '__bytes__', '__repr__',
# '__str__', '__getattr__', '__getattribute__', '__hash__',
# '__setattr__', '__delattr__', '__dir__', '__get__', '__set__',
# '__delete__', '__set_name__',

self_magic_names = ('__str__', '__format__', '__len__', 
                    '__float__', '__invert__', '__complex__', '__int__',
                    '__index__', '__round__', '__trunc__', '__floor__',
                    '__ceil__', '__enter__', '__exit__', '__await__', '__aiter__',
                    '__anext__', '__aenter__', '__aexit__', '__reversed__',
                    '__neg__', '__pos__', '__abs__')

multi_magic_names = ('__call__',)

other_magic_names = ('__lt__', '__le__', '__eq__', '__ne__', 
                      '__gt__', '__ge__', '__bool__', '__getitem__',
                      '__setitem__', '__delitem__', '__missing__',
                      '__contains__', '__add__', '__sub__', '__mul__', '__matmul__',
                      '__truediv__', '__floordiv__', '__mod__', '__divmod__',
                      '__pow__', '__lshift__', '__rshift__', '__and__', '__or__',
                      '__radd__', '__rsub__', '__rmul__', '__rmatmul__',
                      '__rtruediv__', '__rfloordiv__', '__rmod__', '__rdivmod__',
                      '__rpow__', '__rlshift__', '__rrshift__', '__rand__',
                      '__rxor__', '__ror__') 
    
inplace_magic_names = ('__iadd__', '__isub__', '__imul__', '__imatmul__',
                        '__itruediv__', '__idiv__', '__ifloordiv__',
                        '__imod__', '__ipow__', '__ilshift__', '__irshift__',
                        '__iand__', '__ixor__', '__ior__')

class metaProperty(type):
    """Metaclass for FreeProperty and subclasses."""
    
    @property
    def units(cls):
        return cls._units
    
    @units.setter
    def units(cls, units):
        cls._units = units
    
    def getter(cls, fget):
        cls.value = cls.value.getter(fget)
        return cls
    
    def setter(cls, fset):
        cls.value = cls.value.setter(fset)
        return cls


# %% Abstract Property Class

isa = isinstance

class FreeProperty(metaclass=metaProperty):
    """Abstract Property class. Child classes must include a 'value' property."""
    __slots__ = ()
    _units = ''
    
    def __init__(self, *slots):
        set = setattr
        for i, j in zip(self.__slots__, slots): set(self, i, j)
    
    def __call__(self, *args, **kwargs):
        return self.value(*args, **kwargs)
    
    def __len__(self):
        return self.value.__len__()
    
    def __float__(self):
        return self.value.__float__()
    
    def __invert__(self):
        return self.value.__invert__()
    
    def __complex__(self):
        return self.value.__complex__()
    
    def __int__(self):
        return self.value.__int__()
    
    def __index__(self):
        return self.value.__index__()
    
    def __round__(self):
        return self.value.__round__()
    
    def __trunc__(self):
        return self.value.__trunc__()
    
    def __floor__(self):
        return self.value.__floor__()
    
    def __ceil__(self):
        return self.value.__ceil__()
    
    def __enter__(self):
        return self.value.__enter__()
    
    def __exit__(self):
        return self.value.__exit__()
    
    def __await__(self):
        return self.value.__await__()
    
    def __aiter__(self):
        return self.value.__aiter__()
    
    def __anext__(self):
        return self.value.__anext__()
    
    def __aenter__(self):
        return self.value.__aenter__()
    
    def __aexit__(self):
        return self.value.__aexit__()
    
    def __reversed__(self):
        return self.value.__reversed__()
    
    def __neg__(self):
        return self.value.__neg__()
    
    def __pos__(self):
        return self.value.__pos__()
    
    def __abs__(self):
        return self.value.__abs__()
    
    def __bool__(self):
        return self.value.__bool__()
    
    def __format__(self, other):
        return self.value.__format__(other)
    
    def __lt__(self, other):
        return self.value.__lt__(other.value if isa(other, FreeProperty) else other)
    
    def __le__(self, other):
        return self.value.__le__(other.value if isa(other, FreeProperty) else other)
    
    def __eq__(self, other):
        return self.value.__eq__(other.value if isa(other, FreeProperty) else other)
    
    def __ne__(self, other):
        return self.value.__ne__(other.value if isa(other, FreeProperty) else other)
    
    def __gt__(self, other):
        return self.value.__gt__(other.value if isa(other, FreeProperty) else other)
    
    def __ge__(self, other):
        return self.value.__ge__(other.value if isa(other, FreeProperty) else other)
    
    def __getitem__(self, other):
        return self.value.__getitem__(other.value if isa(other, FreeProperty) else other)
    
    def __setitem__(self, other):
        return self.value.__setitem__(other.value if isa(other, FreeProperty) else other)
    
    def __delitem__(self, other):
        return self.value.__delitem__(other.value if isa(other, FreeProperty) else other)
    
    def __missing__(self, other):
        return self.value.__missing__(other.value if isa(other, FreeProperty) else other)
    
    def __contains__(self, other):
        return self.value.__contains__(other.value if isa(other, FreeProperty) else other)
    
    def __add__(self, other):
        return self.value.__add__(other.value if isa(other, FreeProperty) else other)
    
    def __sub__(self, other):
        return self.value.__sub__(other.value if isa(other, FreeProperty) else other)
    
    def __mul__(self, other):
        return self.value.__mul__(other.value if isa(other, FreeProperty) else other)
    
    def __matmul__(self, other):
        return self.value.__matmul__(other.value if isa(other, FreeProperty) else other)
    
    def __truediv__(self, other):
        return self.value.__truediv__(other.value if isa(other, FreeProperty) else other)
    
    def __floordiv__(self, other):
        return self.value.__floordiv__(other.value if isa(other, FreeProperty) else other)
    
    def __mod__(self, other):
        return self.value.__mod__(other.value if isa(other, FreeProperty) else other)
    
    def __divmod__(self, other):
        return self.value.__divmod__(other.value if isa(other, FreeProperty) else other)
    
    def __pow__(self, other):
        return self.value.__pow__(other.value if isa(other, FreeProperty) else other)
    
    def __lshift__(self, other):
        return self.value.__lshift__(other.value if isa(other, FreeProperty) else other)
    
    def __rshift__(self, other):
        return self.value.__rshift__(other.value if isa(other, FreeProperty) else other)
    
    def __and__(self, other):
        return self.value.__and__(other.value if isa(other, FreeProperty) else other)
    
    def __or__(self, other):
        return self.value.__or__(other.value if isa(other, FreeProperty) else other)
    
    def __radd__(self, other):
        return self.value.__radd__(other.value if isa(other, FreeProperty) else other)
    
    def __rsub__(self, other):
        return self.value.__rsub__(other.value if isa(other, FreeProperty) else other)
    
    def __rmul__(self, other):
        return self.value.__rmul__(other.value if isa(other, FreeProperty) else other)
    
    def __rmatmul__(self, other):
        return self.value.__rmatmul__(other.value if isa(other, FreeProperty) else other)
    
    def __rtruediv__(self, other):
        return self.value.__rtruediv__(other.value if isa(other, FreeProperty) else other)
    
    def __rfloordiv__(self, other):
        return self.value.__rfloordiv__(other.value if isa(other, FreeProperty) else other)
    
    def __rmod__(self, other):
        return self.value.__rmod__(other.value if isa(other, FreeProperty) else other)
    
    def __rdivmod__(self, other):
        return self.value.__rdivmod__(other.value if isa(other, FreeProperty) else other)
    
    def __rpow__(self, other):
        return self.value.__rpow__(other.value if isa(other, FreeProperty) else other)
    
    def __rlshift__(self, other):
        return self.value.__rlshift__(other.value if isa(other, FreeProperty) else other)
    
    def __rrshift__(self, other):
        return self.value.__rrshift__(other.value if isa(other, FreeProperty) else other)
    
    def __rand__(self, other):
        return self.value.__rand__(other.value if isa(other, FreeProperty) else other)
    
    def __rxor__(self, other):
        return self.value.__rxor__(other.value if isa(other, FreeProperty) else other)
    
    def __ror__(self, other):
        return self.value.__ror__(other.value if isa(other, FreeProperty) else other)
    
    def __iadd__(self, other):
        self.value = self.value.__add__(other.value if isa(other, FreeProperty) else other)
        return self
    
    def __isub__(self, other): 
        self.value = self.value.__sub__(other.value if isa(other, FreeProperty) else other)
        return self
    
    def __imul__(self, other): 
        self.value = self.value.__mul__(other.value if isa(other, FreeProperty) else other)
        return self
    
    def __imatmul__(self, other):
        self.value = self.value.__matmul__(other.value if isa(other, FreeProperty) else other)
        return self
    
    def __itruediv__(self, other): 
        self.value = self.value.__truediv__(other.value if isa(other, FreeProperty) else other)
        return self
    
    def __ifloordiv__(self, other):
        self.value = self.value.__floordiv__(other.value if isa(other, FreeProperty) else other)
        return self
    
    def __imod__(self, other): 
        self.value = self.value.__mod__(other.value if isa(other, FreeProperty) else other)
        return self
    
    def __ipow__(self, other): 
        self.value = self.value.__pow__(other.value if isa(other, FreeProperty) else other)
        return self
    
    def __ilshift__(self, other):
        self.value = self.value.__lshift__(other.value if isa(other, FreeProperty) else other)
        return self
    
    def __irshift__(self, other):
        self.value = self.value.__rshift__(other.value if isa(other, FreeProperty) else other)
        return self
    
    def __iand__(self, other): 
        self.value = self.value.__and__(other.value if isa(other, FreeProperty) else other)
        return self
    
    def __ixor__(self, other): 
        self.value = self.value.__xor__(other.value if isa(other, FreeProperty) else other)
        return self
    
    def __ior__(self, other):
        self.value = self.value.__or__(other.value if isa(other, FreeProperty) else other)
        return self
    
    def __repr__(self):
        units = self._units
        if units: units = f' {units}'
        value = self.value
        try: value = format(value, '.5g')
        except: pass
        return f'<{self.name}: {value}{units}>'
    
    def show(self):
        units = self._units
        if units: units = f' {units}'
        value = self.value
        try: value = format(value, '.5g')
        except: pass
        print(f'<{type(self).__name__}({self.name}): {value}{units}>')
    
    _ipython_display_ = show
    
    


# -*- coding: utf-8 -*-
"""
Created on Thu Jan 10 03:56:02 2019

@author: yoelr
"""
import numpy as np

__all__ = ('property_array',)

ndarray = np.ndarray
broadcast = np.broadcast_to
isa = isinstance

# %% Property array

class property_array(ndarray):
    """Create an array that allows for array-like manipulation of FreeProperty objects. All entries in a property_array must be instances of FreeProperty. Setting items of a property_array sets values of objects instead.
    
    Parameters
    ----------
        array : array_like[FreeProperty]
            Input data, in any form that can be converted to an array. This includes lists, lists of tuples, tuples, tuples of tuples, tuples of lists and ndarrays.
    
        order : {'C', 'F'}
            Whether to use row-major (C-style) or column-major (Fortran-style) memory representation. Defaults to ‘C’.
    
    Examples
    --------
        Use the PropertyFactory to create a Weight property class which calculates weight based on density and volume:
    
        .. code-block:: python
        
            >>> from free_property import PropertyFactory, property_array
            >>>
            >>> @PropertyFactory
            >>> def Weight(self):
            ...    '''Weight (kg) based on volume (m^3).'''
            ...    data = self.data
            ...    rho = data['rho'] # Density (kg/m^3)
            ...    vol = data['vol'] # Volume (m^3)
            ...    return rho * vol
            >>>
            >>> @Weight.setter
            >>> def Weight(self, weight):
            ...    data = self.data
            ...    rho = data['rho'] # Density (kg/m^3)
            ...    data['vol'] = weight / rho
           
        Create dictionaries of data and initialize new properties:
       
        .. code-block:: python
       
           >>> water_data = {'rho': 1000, 'vol': 3}
           >>> ethanol_data = {'rho': 789, 'vol': 3}
           >>> weight_water = Weight('Water', water_data)
           >>> weight_ethanol = Weight('Ethanol', ethanol_data)
           >>> weight_water
           <Weight(Water): 3000 kg>
           >>> weight_ethanol
           <Weight(Ethanol): 2367 kg>
          
        Create a property_array from data:
           
        .. code-block:: python
       
           >>> prop_arr = property_array([weight_water, weight_ethanol])
           >>> prop_arr
           property_array([<Water: 3000 kg>, <Ethanol: 2367 kg>], dtype=object)
           
        Changing the values of a property_array changes the value of its properties:
           
        .. code-block:: python
       
           >>> # Addition in place
           >>> prop_arr += 3000
           >>> prop_arr
           property_array([<Water: 6000 kg>, <Ethanol: 5367 kg>], dtype=object)
           >>> # Note how the data also changes
           >>> water_data
           {'rho': 1000, 'vol': 6.0}
           >>> ethanol_data
           {'rho': 789, 'vol': 6.802281368821292}
           >>> # Setting an item changes the property value
           >>> prop_arr[1] = 2367
           >>> ethanol_data
           {'rho': 789, 'vol': 3}
          
        New arrays have no connection to the property_array:
           
        .. code-block:: python
       
           >>> prop_arr - 1000 #  Returns a new array
           array([5000.0, 1367.0], dtype=object)
           >>> water_data #  Data remains unchanged
           {'rho': 1000, 'vol': 6.0}
    
    """
    __slots__ = ()          
    
    def __new__(cls, properties, order='C'):
        return np.asarray(properties, object, order).view(cls)
    
    def __setitem__(self, key, value):
        items = self[key]
        if isa(items, ndarray):
            for i, v in np.nditer((items, value), flags=('refs_ok',)):
                i.item().value = v 
        else:
            items.value = value
    
    def __array_wrap__(self, result):
        if self is result: return self
        else: return result.view(ndarray)
    
    def __iadd__(self, other):
        self[:] = ndarray.__add__(self, other)
        return self
    
    def __isub__(self, other): 
        self[:] = ndarray.__sub__(self, other)
        return self
    
    def __imul__(self, other): 
        self[:] = ndarray.__mul__(self, other)
        return self
    
    def __imatmul__(self, other):
        raise TypeError("in-place matrix multiplication is not (yet) supported")
    
    def __itruediv__(self, other): 
        self[:] = ndarray.__truediv__(self, other)
        return self
    
    def __ifloordiv__(self, other):
        self[:] = ndarray.__floordiv__(self, other)
        return self
    
    def __imod__(self, other): 
        self[:] = ndarray.__mod__(self, other)
        return self
    
    def __ipow__(self, other): 
        self[:] = ndarray.__pow__(self, other)
        return self
    
    def __ilshift__(self, other):
        self[:] = ndarray.__lshift__(self, other)
        return self
    
    def __irshift__(self, other):
        self[:] = ndarray.__rshift__(self, other)
        return self
    
    def __iand__(self, other): 
        self[:] = ndarray.__and__(self, other)
        return self
    
    def __ixor__(self, other): 
        self[:] = ndarray.__xor__(self, other)
        return self
    
    def __ior__(self, other):
        self[:] = ndarray.__or__(self, other)
        return self




# -*- coding: utf-8 -*-
"""
Created on Fri Jan 18 14:20:29 2019

@author: Guest Group
"""

__all__ = ('FreeProperty' , 'PropertyFactory', 'property_array')

from ._free_property import FreeProperty
from ._property_factory import PropertyFactory
from ._property_array import property_array


# -*- coding: utf-8 -*-
"""
Created on Fri Jan 18 14:18:27 2019

@author: Guest Group
"""
from ._free_property import metaProperty, FreeProperty

__all__ = ('PropertyFactory',)

# %% utils

def search_units(doc):
    # Find units
    par1 = doc.find('(')
    par2 = doc.find(')')
    if par1==-1 or par2==-1:
        units = ''
    else:
        units = doc[par1+1:par2]
    return units


# %% Property Factory


def PropertyFactory(fget=None, fset=None, clsname=None, doc=None, units=None,
                    slots=None):
    """Create an FreeProperty subclass with getter and setter functions.
    
    Parameters
    ----------
    
        fget : function
            Should return value of instances.
        
        fset : function
            Should set the value of instances.
        
        clsname : str
            Name of the class. If None, the function name of fget will be used
        
        doc : str
            Docstring of class. If None, the docstring of fget will be used.
        
        units : str
            Units of measure. If None, the docstring will be searched and units of measure in parenthesis will be used.
    
        slots : tuple[str]
            Slots for class.
    
    Examples
    --------
    
    The PropertyFactory is a FreeProperty class creator that functions similar to Python 'property' objects. Use the PropertyFactory to create a Weight class which calculates weight based on density and volume:
    
    .. code-block:: python
        
        >>> def getter(self):
        ...    '''Weight (kg) based on volume (m^3).'''
        ...    data = self.data
        ...    rho = data['rho'] # Density (kg/m^3)
        ...    vol = data['vol'] # Volume (m^3)
        ...    return rho * vol
        >>>
        >>> def setter(self, weight):
        ...    data = self.data
        ...    rho = data['rho'] # Density (kg/m^3)
        ...    data['vol'] = weight / rho
        >>>
        >>> # Initialize with a value getter, setter, and the class name.
        >>> Weight = PropertyFactory(getter, setter, 'Weight')
        
    It is more convinient to use the PropertyFactory as a decorator:
    
    .. code-block:: python
       
        >>> @PropertyFactory
        >>> def Weight(self):
        ...    '''Weight (kg) based on volume (m^3).'''
        ...    data = self.data
        ...    rho = data['rho'] # Density (kg/m^3)
        ...    vol = data['vol'] # Volume (m^3)
        ...    return rho * vol
        >>>
        >>> @Weight.setter
        >>> def Weight(self, weight):
        ...    data = self.data
        ...    rho = data['rho'] # Density (kg/m^3)
        ...    data['vol'] = weight / rho
       
    
    Create dictionaries of data and initialize new Weight objects:
   
    .. code-block:: python
   
       >>> water_data = {'rho': 1000, 'vol': 3}
       >>> ethanol_data = {'rho': 789, 'vol': 3}
       >>> weight_water = Weight('Water', water_data)
       >>> weight_ethanol = Weight('Ethanol', ethanol_data)
       >>> weight_water
       Weight(Water): 3000 kg
       >>> weight_ethanol
       Weight(Ethanol): 2367 kg
    
    .. Note::
        
       The units are taken from the the function docstring. The first word in parentesis denotes the units.
    
    These properties behave just like their dynamic value:
    
    .. code-block:: python
    
        >>> weight_water + 30
        3030
        >>> weight_water + weight_ethanol
        5367
        
    Get and set the value through the 'value' attribute:
        
    .. code-block:: python
    
        >>> weight_water.value
        3000
        >>> weight_water.value = 4000 
        >>> weight_water.value
        4000
        >>> water_data # Note that the volume changed too
        {'rho': 1000, 'vol': 4}

    In place magic methods will also change the property value:
    
    .. code-block:: python
    
        >>> weight_water -= 1000
        >>> weight_water
        Weight(Water): 3000 kg
        >>> water_data  # The change also affects the original data
        {'rho': 1000, 'vol': 3}
            
    """       
    if not fget:
        return lambda fget: PropertyFactory(fget, fset, clsname, doc, units, slots)
        
    # Defaults
    if clsname is None: clsname = fget.__name__
    if doc     is None: doc     = fget.__doc__
    if units   is None: units   = search_units(doc)
    
    definitions = {'__doc__': doc,
                   '__slots__': slots or ('name', 'data'),
                   '_units': units,
                   'value': property(fget, fset)}        
    return metaProperty(clsname, (FreeProperty,), definitions)
